/*
 * mupdate: an area file converter, translates MercII/Diku --> Circle
 * Usage: mupdate [-v] mercarea.are [...]
 *    or: mupdate [-v] -m dikuarea.mob out.mob
 *
 * Takes a MercII .are-type area file, splits it up into it's component
 * parts, and converts it to .obj, .mob, .wld, and .zon files with
 * Circle-style info.  Doesn't handle #SHOPS sections yet.  Creates
 * a boring .spe text file with all spec_procs.  Appends any info
 * found in a #HELPS section to one main help file.  Doesn't recover
 * from input errors very well.  Warnings vary in severity, but may
 * be helpful once in a blue moon.
 *
 * Also has a -m mode: mupdate -m in.mob out.mob converts an early
 * Diku-style .mob file into Circle-style output.
 * 
 * Read the man page!
 * 
 * Feel free to email me comments -- no guarantees that I will
 * support it, though.  [Maybe Jeremy Elson is willing?  Dunno.]
 * 
 * David Wagner dawagner@princeton.edu
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>

#define DEFAULT_LIFESPAN	30	/* minutes */
#define DEFAULT_RESETMODE	2	/* reset even if area not deserted */
#define HELPFILE		"help"	/* output file for #HELPS stuff */
#define	LINEIN	if (fgets(line, sizeof(line), in) == NULL) \
			die("Unexpected EOF!");

FILE	*in=NULL;			/* the current input stream */
int	zonenum=-1;			/* this zone's number */
char	zonename[160];			/* this zone's name */
int	zonelastroom=-1;		/* this zone's last room */
char	line[256];			/* the current line */
int	helpfilegrown=0;		/* was anything added to HELPFILE? */
int	verbose=0;			/* spam the user with warnings? */

/* Youch!  Hit a fatal error, let's get outta here! */
void	die(m)
	char	*m;
{
	fprintf(stderr, "FATAL ERROR:\t%s\n", m);
	exit(1);
}

/*
 * Concatenates m1 and m2 to produce a warning message, which is printed
 * to the screen.  Collects warnings of similar types to reduce screen
 * spam, so that the following calls:
 *
 *	warn("ignoring ", "unimplemented shop command");
 *	warn("ignoring ", "strange input line");
 *	warn("ignoring ", "obnoxious usage examples");
 *	warn("out of memory", "");
 *
 * produce this output:
 *
 *	WARNING:	ignoring unimplemented shop command.
 *	WARNING:	last message type repeated 2 more times.
 *	WARNING:	out of memory.
 *
 * Also, the special call warn("", "") flushes all saved up warnings.
 * 
 * Finally, if verbose is set (see -v flag), no warning buffering
 * or deletion is done.
 */
void	warn(m1, m2)
	char	*m1;
	char	*m2;
{
	static char	last[160];
	static int	firstwarning=1, repeats=0;
	char		msg[160];

	if (verbose) {
		strcpy(msg, m1);
		if (strchr(msg, '\n'))
			*(strchr(msg, '\n')) = '\0';
		strcat(msg, m2);
		if (strchr(msg, '\n'))
			*(strchr(msg, '\n')) = '\0';
		fprintf(stderr, "%s\n", msg);
		return;
	}

	if (*m1 == '\0' && *m2 == '\0') {
		if (repeats > 0)
			fprintf(stderr, "WARNING:\tlast message type repeated %d more times.\n", repeats);
		repeats = 0;
		last[0] = '\0';
		return;
	}
	strcpy(msg, m1);
	if (strchr(msg, '\n'))
		*(strchr(msg, '\n')) = '\0';
	strcat(msg, m2);
	if (strchr(msg, '\n'))
		*(strchr(msg, '\n')) = '\0';

	if (firstwarning)
		firstwarning = 0;
	else if (strcmp(last, m1) == 0) {
		repeats++;
		return;
	}

	if (repeats > 0)
		fprintf(stderr, "WARNING:\tlast message type repeated %d more times.\n", repeats);
	repeats = 0;
	fprintf(stderr, "WARNING:\t%s.\n", msg);
	strcpy(last, m1);
}

/* Similar to atoi(), except that mercatoi("1|2|64") returns 67. */
int	mercatoi(s)
	char	*s;
{
	int	n;
	char	*p;

	for (n=0; p=strchr(s, '|'); s=p) {
		*p++ = '\0';
		n += atoi(s);
	}
	return(n+atoi(s));
}

/*
 * Copies a Diku string [text terminated by a tilde] from stdin
 * to the named output stream.  If the addnl flag is set, it
 * converts lines like:
 *
 *	A green slime mold is here.  It stinks like sh**.~
 *
 * into somethign like this:
 *
 *	A green slime mold is here.  It stinks like sh**.
 *	~
 */
void	copystr(out, addnl)
	FILE	*out;
	int	addnl;
{
	for (;;) {
		LINEIN
		if (strchr(line, '~')) {
			if (line[0] == '~' || !addnl)
				fputs(line, out);
			else {
				*(strchr(line, '~')) = '\0';
				fprintf(out, "%s\n~\n", line);
			}
			break;
		}
		fputs(line, out);
	}
}

/*
 * Processes the #HELPS section.  Just appends everything in the section
 * onto the output file HELPFILE.
 */
void	helps( /* void */ )
{
	int	level;
	char	*s;
	FILE	*out;

	if ((out=fopen(HELPFILE, "a")) == NULL)
		die("couldn't open output file.");

	for (;;) {
		LINEIN
		if ((s=strtok(line, " \t\n")) == NULL)
			continue;
		helpfilegrown = 1;
		level = mercatoi(s);
		s = strtok(NULL, " \t\n");
		if (level==0 && strcmp(s, "$~")==0)
			break;
		fprintf(out, "%d\t%s\n", level, s);
		copystr(out, 0);
	}

	fclose(out);
}

/*
 * Processes the #MOBILES section of the .are file [or alternatively
 * processes the input file when mupdate is run with the -m flag].
 * Stops processing when it sees a line of the form:
 *
 *	#<lastnum>
 *
 * Normally, this function is run with lastnum=0 [but when mupdate is
 * run with the -m flag, lastnum=99999].
 *
 * Does all sorts of conversions on the input stream.  Can only handle
 * simple mobs right now, but that shouldn't be a problem, as I haven't
 * seen even ONE monster that doesn't have the simple flag set.  Adds
 * multiple attacks to each monster based its level.  This may need some
 * finetuning by hand after the program is run.
 */
void	mobiles(lastnum)
	int	lastnum;
{
	int	attacks, thac0, level, n;
	char	*p, damage[40];
	FILE	*out;

	if ((out=fopen("tmp.mob", "w")) == NULL)
		die("couldn't open output file.");

	for (;;) {
		/* #<virtual number>	->	#<virtual number> */
		LINEIN
		if (strcspn(line, " \t\n") == 0)
			continue;
		if (line[0] == '>') {
			warn("ignoring spec proc: ", line);
			for (;;) {
				LINEIN
				if (line[0] == '|' || line[0] == '#')
					break;
				warn("ignoring spec proc: ", line);
			}
			if (line[0] != '#')
				continue;
			warn("woops! ignored too much of the spec proc?", "");
		}
		if (line[0] != '#' || !isdigit(line[1])) {
			warn("ignoring input line: ", line);
			continue;
		}
		if ((n=mercatoi(line+1)) == lastnum)
			break;
		fprintf(out, "#%d\n", n);
		/* <namelist>~		->	#<namelist>~ */
		LINEIN
		fputs(line, out);
		/* <short description>~ ->	#<short description>~ */
		LINEIN
		if (strchr(line, '~') == NULL) {
			fputs(line, out);
			LINEIN
			if (strchr(line, '~') == NULL)
				die("obnoxiously long short description");
		}
		fputs(line, out);
		/* <long description>~ ->	#<long description><NL>~ */
		copystr(out,1);
		/* <description>~	->	#<description><NL>~ */
		copystr(out, 1);
		/*
		 * <action> <affection> <alignment> <detailed/simple>
		 *			->
		 * <action> <affection> <alignment> M<NL>
		 *
		 * BUGBUG: I just copy over the action flag!!
		 * BUGBUG: I can only handle Simple mobs right now!!
		 */
		LINEIN
		if (strchr(line, '~')) {
			/* stupid buggy non-compliant areas <sigh> */
			warn("ignoring input line: ", line);
			LINEIN
		}
		fprintf(out, "%d ", mercatoi(strtok(line, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		p = strtok(NULL, " \t");
		switch(*p) {
			case 'S': fprintf(out, "M\n"); break;;
			case 'M': die("Merc mob found!");
			case 'D': die("Detailed mob found!");
			default:  die("Unknown simple-flag!");
		}
		/*
		 * <level> <thac0> <AC> <hp> <damage><NL>
		 * <gold> <exp><NL>
		 * <position> <default position> <sex>
		 *			->
		 * <level> <AC> <num attacks> <hp> <gold><NL>
		 * <% chance> <thac0> <damage> <attack type><NL> ...
		 * <exp> <position> <default position> <sex><NL>
		 */
		LINEIN
		level = mercatoi(strtok(line, " \t"));
		if (level <= 10)
			attacks = 1;
		else if (level <= 15)
			attacks = 2;
		else if (level <= 20)
			attacks = 3;
		else
			attacks = 4;
		thac0 = mercatoi(strtok(NULL, " \t"));
		fprintf(out, "%d %d %d ", level, mercatoi(strtok(NULL, " \t")),
					attacks);
		if ((p = strtok(NULL, " \t")) == NULL) {
			LINEIN
			if ((p = strtok(line, " \t")) == NULL)
				die("really obnoxious lack of hp field.");
		}
		fputs(p, out);
		if ((p = strtok(NULL, " \t")) == NULL) {
			LINEIN
			if ((p = strtok(line, " \t")) == NULL)
				die("really obnoxious lack of damage field.");
		}
		strcpy(damage, p);
		LINEIN
		fprintf(out, " %dd1\n", mercatoi(strtok(line, " \t")));
		while (attacks-- > 0)
			fprintf(out, "100 %d %s 0\n", thac0, damage);
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t\n")));
		LINEIN
		fprintf(out, "%d ", mercatoi(strtok(line, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d\n", mercatoi(strtok(NULL, " \t\n")));
	}

	fprintf(out, "#99999\n$~\n");
	fclose(out);
}

/* Processes the #OBJECTS section.  Does only minor changes. */
void	objects( /* void */ )
{
	int	c, n;
	FILE	*out;

	if ((out=fopen("tmp.obj", "w")) == NULL)
		die("couldn't open output file.");

	for (;;) {
		LINEIN
		if (strcspn(line, " \t\n") == 0)
			continue;
		if (line[0] != '#' || !isdigit(line[1])) {
			warn("ignoring input line: ", line);
			continue;
		}
		if ((n=mercatoi(line+1)) == 0)
			break;
		fprintf(out, "#%d\n", n);
		/* <namelist>~		->	#<namelist>~ */
		LINEIN
		fputs(line, out);
		/* <short description>~ ->	#<short description>~ */
		LINEIN
		if (strchr(line, '~') == NULL) {
			fputs(line, out);
			LINEIN
			if (strchr(line, '~') == NULL)
				die("obnoxiously long short description");
		}
		fputs(line, out);
		/* <long description>~	->	#<long description><NL>~ */
		copystr(out, 0);
		/* <action description>~ ->	#<action description><NL>~ */
		copystr(out, 1);
		/*
		 * <item-type> <extra-flags> <wear-flags>
		 *			->
		 * <item-type> <extra-flags> <wear-flags>
		 */
		LINEIN
		fprintf(out, "%d ", mercatoi(strtok(line, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d\n", mercatoi(strtok(NULL, " \t\n")));
		/*
		 * <value0> <value1> <value2> <value3>
		 *			->
		 * <value0> <value1> <value2> <value3>
		 */
		LINEIN
		fprintf(out, "%d ", mercatoi(strtok(line, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d\n", mercatoi(strtok(NULL, " \t\n")));
		/*
		 * <weight> <cost> <cost/day>
		 *			->
		 * <weight> <cost> <cost/day>
		 */
		LINEIN
		fprintf(out, "%d ", mercatoi(strtok(line, " \t")));
		fprintf(out, "%d ", mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d\n", mercatoi(strtok(NULL, " \t\n")));
		/*
		 * E/A sections
		 */
		for (;;) {
			c = getc(in);
			if (c!='E' && c!='A') {
				ungetc(c, in);
				break;
			}
			LINEIN
			if (c=='E') {
				fputs("E\n", out);

				/* hack to check for non-compliant areas */
				if (strchr(line, '~')) {
					/* goddamned annoying shitty areas */
					warn("E must be followed by newline, "
						"damnit! fix by hand: ", line);
					fprintf(out, "%s", line);
				} else
					copystr(out, 0);
				copystr(out, 0);
			} else {
				LINEIN
				fprintf(out, "A %d ",
					mercatoi(strtok(line, " \t")));
				fprintf(out, "%d\n",
					mercatoi(strtok(NULL, " \t\n")));
			}
		}
	}

	fprintf(out, "$~\n");
	fclose(out);
}

/*
 * Processes the #ROOMS section.  Sets all kinds of global variables,
 * especially the ones for later use by the resets() function.
 * There's all sorts of wierd-ities that can happen, since MercII
 * mud ignores certain zone information fields and uses some info
 * of it's own, so I haveta wing it sometimes...  Here's where a
 * lot of the strange warnings will come from.
 */
void	rooms( /* void */ )
{
	int	n;
	char	*p;
	FILE	*out;

	zonenum = -1;
	zonelastroom = -1;
	if ((out=fopen("tmp.wld", "w")) == NULL)
		die("couldn't open output file.");

	for (;;) {
		/* #<virtual number>	->	#<virtual number> */
		LINEIN
		if (strcspn(line, " \t\n") == 0)
			continue;
		if (line[0] != '#' || !isdigit(line[1])) {
			warn("ignoring input line: ", line);
			continue;
		}
		if ((n=mercatoi(line+1)) == 0)
			break;
		if (zonenum < 0) {
			zonenum = n/100;
			printf("Updating zone #%d: `%s'.\n", zonenum, zonename);
		} else if (n/100 != zonenum)
			/*
			 * BUGBUG: two room numbers in any one zone can only
			 * differ by their last two digits.
			 */
			warn("zone mismatch: ", "partitioning problem");
		if (n > zonelastroom)
			zonelastroom = n;
		fprintf(out, "#%d\n", n);
		/* <namelist>~		->	#<namelist>~ */
		LINEIN
		fputs(line, out);
		/* <description>~	->	#<description>~ */
		copystr(out, 1);
		/* <area> <flags> <sector> ->	<area> <flags> <sector> */
		LINEIN
		n = mercatoi(strtok(line, " \t"));
		if (n!=zonenum) {
			/*
			 * Youch!  The zone info listed in the input
			 * looks wrong.  Of course, that's not too too
			 * surprising, since MercII ignores that field...
			 */
			char	m[40];

			sprintf(m, "%d != %d", n, zonenum);
			warn("zone mismatch: ", m);
		}
		fprintf(out, "%d %d ", zonenum, mercatoi(strtok(NULL, " \t")));
		fprintf(out, "%d\n", mercatoi(strtok(NULL, " \t\n")));
		/* D/E sections */
		for (;;) {
			LINEIN
			if (line[0]=='S')
				break;
			if (line[0]=='D') {
				for (p=line+1; *p==' ' || *p=='\t'; p++)
					;
				fprintf(out, "D%d\n", mercatoi(p));
				copystr(out, 0);
				copystr(out, 0);
				LINEIN
				fprintf(out, "%d ",
					mercatoi(strtok(line, " \t")));
				fprintf(out, "%d ",
					mercatoi(strtok(NULL, " \t")));
				fprintf(out, "%d\n",
					mercatoi(strtok(NULL, " \t\n")));
			} else if (line[0]=='E') {
				fputs(line, out);
				copystr(out, 0);
				copystr(out, 0);
			} else
				warn("ignoring input line: ", line);
		}
	}

	fprintf(out, "#99999\n$~\n");
	fclose(out);
}

/*
 * Processes the #RESETS section and creates a .zon file from
 * the information found there and from the information stored
 * in the global variables (which were set from junk in the #ROOMS
 * section).  Also massages the zone name a little, since the
 * MercII convention is to use a certain special (annoying) format.
 */
void	resets( /* void */ )
{
	char	*p;
	FILE	*out;

	if ((out=fopen("tmp.zon", "w")) == NULL)
		die("couldn't open output file.");

	if (zonenum < 0 || zonelastroom < 0) {
		/*
		 * What bozo would try to convert proto.are?
		 * [Besides me, that is: I like to do 'mupdate *.are'
		 * in the merc area/ directory for testing.]
		 */
		warn("if you were running me on proto.are, shame on you", "");
		warn("or maybe the #RESETS section came before #ROOMS?!", "");
		die("#RESETS found before any rooms were defined.");
	}
	fprintf(out, "#%d\n", zonenum);
	if (zonename[0] == '{' && (p=strchr(zonename, '}')) != NULL) {
		for (p++; *p==' ' || *p=='\t'; p++)
			;
		for (; *p && *p!=' ' && *p!='\t'; p++)
			;
		if (*p == '\0')
			die("wierd zone name!");
		*p++ = '\0';
		for (; *p==' ' || *p=='\t'; p++)
			;
		fprintf(out, "%s\t[%s]~\n", p, zonename);
	} else
		fprintf(out, "%s~\n", zonename);
	fprintf(out, "%d ", zonelastroom);
	fprintf(out, "%d ", DEFAULT_LIFESPAN);
	fprintf(out, "%d\n", DEFAULT_RESETMODE);

	for (;;) {
		LINEIN
		if (line[0] == 'R') {
			warn("ignoring untranslateable .zon command: ", line);
			continue;
		}
		fputs(line, out);
		if (line[0] == 'S')
			break;
	}

	fprintf(out, "#99999\n$~\n");
	fclose(out);
}

/* Copies info listed in the #SPECIALS section to an output file. */
void	specials( /* void */ )
{
	FILE	*out;

	if ((out=fopen("tmp.spe", "a")) == NULL)
		die("couldn't open output file.");

	for (;;) {
		LINEIN
		if (line[0] == 'S')
			break;
		fputs(line, out);
	}

	fprintf(out, "#99999\n$~\n");
	fclose(out);
}

/* Copy spec-proc information (for annoying non-compliant .are files) */
void	mobprog( /* void */ )
{
	specials();
}

/* Does all the real work in processing one .are [ie MercII area] file. */
void	convertare( /* void */ )
{
	char	zs[80], *p;

	(void) unlink("tmp.hlp");
	(void) unlink("tmp.mob");
	(void) unlink("tmp.obj");
	(void) unlink("tmp.wld");
	(void) unlink("tmp.zon");
	(void) unlink("tmp.spe");
	zonenum = -1;
	zonelastroom = -1;
	for (;;) {
		LINEIN
		if (line[0] != '#') {
			for (p=line+1; *p==' ' || *p=='\t'; p++)
				;
			if (*p)
				warn("ignoring input line: ", line);
			continue;
		}
		p = line+1;
		if (strstr(p, "AREA") == p) {
			(void) strtok(p, " \t");
			strcpy(zonename, strtok(NULL, "~\n"));
		} else if (strcmp(p, "HELPS\n") == 0)
			helps();
		/* else if (strcmp(p, "MOBILES\n") == 0) */
		/*
		 * hack to deal with buggy #MOBOLD lines, etc.
		 * but ignore #MOBPROG sections!!
		 * argh i hate non-compliant .are files...
		 */
		else if (strstr(p, "MOBPROG") == p)
			mobprog();
		else if (strstr(p, "MOB") == p)
			mobiles(0);
		/* else if (strcmp(p, "OBJECTS\n") == 0) */
		/* hack to deal with buggy #OBJOLD lines, etc. */
		else if (strstr(p, "OBJ") == p)
			objects();
		else if (strcmp(p, "ROOMS\n") == 0)
			rooms();
		else if (strcmp(p, "RESETS\n") == 0)
			resets();
		else if (strcmp(p, "SHOPS\n") == 0)
			for (;;) {
				LINEIN
				if (line[0] == '0' && mercatoi(line) == 0)
					break;
				warn("ignoring shop command: ", line);
			}
		else if (strcmp(p, "SPECIALS\n") == 0)
			specials();
		else if (strcmp(p, "$\n") == 0)
			break;
	}

	if (zonenum < 0) {
		warn("bad zone number: ", "no #ROOMS section, probably");
		return;
	}
	/*
	sprintf(zs, "%d", zonenum);
	(void) rename("tmp.hlp", strcat(zs, ".hlp"));
	*/
	sprintf(zs, "%d", zonenum);
	if (rename("tmp.mob", strcat(zs, ".mob"))) {
		warn("rename tmp.mob failed: ", "no mobiles in area!?");
	}
	sprintf(zs, "%d", zonenum);
	if (rename("tmp.obj", strcat(zs, ".obj"))) {
		warn("rename tmp.obj failed: ", "no objects in area!?");
	}
	sprintf(zs, "%d", zonenum);
	if (rename("tmp.wld", strcat(zs, ".wld")))
		die("rename tmp.wld failed.");
	sprintf(zs, "%d", zonenum);
	if (rename("tmp.zon", strcat(zs, ".zon")))
		die("rename tmp.zon failed.");
	sprintf(zs, "%d", zonenum);
	(void) rename("tmp.spe", strcat(zs, ".spe"));
	warn("", "");
}

void	usage( /* void */ )
{
	fprintf(stderr, "USAGE:\tmupdate [-v] file.are [...]\n");
	fprintf(stderr, "or\tmupdate [-v] -m in.mob out.mob\n");
	exit(1);
}

int	main(argc, argv)
	int	argc;
	char	**argv;
{
	int	i;

	if (argc < 2)
		usage();
	if (strcmp(argv[1], "-v") == 0) {
		verbose = 1;
		argc--;
		argv++;
		if (argc < 2)
			usage();
	}
	if (strcmp(argv[1], "-m") == 0) {
		if (argc!=4)
			usage();
		if ((in=fopen(argv[2], "r")) == NULL)
			die("couldn't open input file.");
		mobiles(99999);
		if (rename("tmp.mob", argv[3]))
			die("rename() failed.");
		return(0);
	}

	for (i=1; i<argc; i++) {
		if ((in=fopen(argv[i], "r")) == NULL)
			die("couldn't open input file.");
		convertare();
		fclose(in);
	}
	fprintf(stderr, "[All zone reset flags set to defaults!]\n");
	if (helpfilegrown)
		fprintf(stderr, "[Text added to output help file `%s'.]\n",
				HELPFILE);
	return(0);
}
